#include "FancyDebug.h"

#include <File.h>
#include <FindDirectory.h>
#include <Path.h>

#include <strstream>

#include "FancyApp.h"
#include "FancyPlayer.h"
#include "FancyPlayList.h"
#include "MessageWindow.h"
#include "Opcodes.h"
#include "PlayStatusView.h"
#include "PlayStatusWindow.h"
#include "VolumeWindow.h"

#define SECONDS(x) x##000000

static const char *kStateFile = "FancyKeyboardPlayer";
static const bigtime_t kShutdownTimeout = SECONDS(60);
static const bigtime_t kPulseRate = SECONDS(15);

#undef SECONDS

FancyApp::FancyApp()
	:	BApplication(kFancyAppSignature)
{	ENTRY

	fMsgWin = new MessageWindow;
	fVolumeWin = new VolumeWindow;
	fPlayer = new FancyPlayer("Player");
	fPlayer->Run();
	fPlayList = new FancyPlayList;

	fPlayWindow = new PlayStatusWindow(fPlayer);

	BMessage m(FPInterface::RegisterCallback);
	if (m.AddMessenger("messenger", BMessenger(this)) != B_OK) FatalError("Unable to add messenger");
	BMessage reply;
	if (BMessenger(fPlayer).SendMessage(&m, &reply) != B_OK) FatalError("Failed to send watch message to player");
	if (reply.what != B_OK) FatalError("Failed to watch player status");

	LoadState();
	Action();
}

void FancyApp::Pulse()
{
	if ((system_time() - fLastAction) < kShutdownTimeout)
		return;

	BMessage reply;
	BMessenger(fPlayer).SendMessage(FPInterface::GetStatus, &reply);

	if (reply.what != FPInterface::Empty)
		return Action();
	else
		be_app->PostMessage(B_QUIT_REQUESTED);
}

void FancyApp::SaveState()
{
	BPath path;
	if (find_directory(B_USER_SETTINGS_DIRECTORY, &path) != B_OK)
		return;
	path.Append(kStateFile);

	BFile file(path.Path(), B_CREATE_FILE | B_ERASE_FILE | B_READ_WRITE);
	if (file.InitCheck() != B_OK)
		return;

	BMessage msg;
	for (FancyPlayList::entry_vec::const_iterator i = fPlayList->GetList().begin(); i != fPlayList->GetList().end(); ++i)
		msg.AddRef("refs", &*i);
	msg.AddRef("currentRef", &fCurrentRef);

	msg.Flatten(&file);
}

void FancyApp::LoadState()
{
	BPath path;
	if (find_directory(B_USER_SETTINGS_DIRECTORY, &path) != B_OK)
		return;
	path.Append(kStateFile);

	BFile file(path.Path(), B_READ_ONLY);
	if (file.InitCheck() != B_OK)
		return;

	BMessage msg;
	if (msg.Unflatten(&file) != B_OK)
		return;

	StopPlaying();
	fPlayList->Empty();
	AddRefs(&msg, "refs");
	msg.FindRef("currentRef", &fCurrentRef);
}

FancyApp::~FancyApp()
{	DENTRY

	SaveState();

	fPlayer->Lock();
	fPlayer->Quit();
	delete fPlayList;
}

void FancyApp::MessageReceived(BMessage *msg)
{	ENTRY

	switch (msg->what) {
	case opcode::Mute:
	case opcode::VolumeUp:
	case opcode::VolumeDown:
		Action();
		fVolumeWin->PostMessage(msg);
		break;

	case opcode::FirstTrack:
		{
			Action();
			if (IsPaused())
				StopPlaying();
			entry_ref first;
			if (fPlayList->GetFirst(first)) {
				if (first != fCurrentRef)
					SetCurrent(first, false);
			} else
				return fMsgWin->ShowText("playlist is empty");
			BMessage m(PlayStatusView::BogusStatus);
			m.AddRef("ref", &fCurrentRef);
			fPlayWindow->PostMessage(&m);
		}
		break;
	case opcode::LastTrack:
		{
			Action();
			if (IsPaused())
				StopPlaying();
			entry_ref last;
			if (fPlayList->GetLast(last)) {
				if (last != fCurrentRef)
					SetCurrent(last, false);
			} else
				return fMsgWin->ShowText("playlist is empty");
			BMessage m(PlayStatusView::BogusStatus);
			m.AddRef("ref", &fCurrentRef);
			fPlayWindow->PostMessage(&m);
		}
		break;
	case opcode::PrevTrack:
		{
			Action();
			if (IsPaused())
				StopPlaying();
			entry_ref prev;
			if (fPlayList->GetPrevious(fCurrentRef, prev))
				SetCurrent(prev, false);
			else if (!fPlayList->ContainsFile(fCurrentRef)) {
				if (fPlayList->GetLast(prev))
					SetCurrent(prev, false);
				else
					return fMsgWin->ShowText("playlist is empty");
			}
			BMessage m(PlayStatusView::BogusStatus);
			m.AddRef("ref", &fCurrentRef);
			fPlayWindow->PostMessage(&m);
		}
		break;
	case opcode::NextTrack:
		{
			Action();
			if (IsPaused())
				StopPlaying();
			entry_ref next;
			if (fPlayList->GetNext(fCurrentRef, next))
				SetCurrent(next, false);
			else if (!fPlayList->ContainsFile(fCurrentRef)) {
				if (fPlayList->GetFirst(next))
					SetCurrent(next, false);
				else
					return fMsgWin->ShowText("playlist is empty");
			}
			BMessage m(PlayStatusView::BogusStatus);
			m.AddRef("ref", &fCurrentRef);
			fPlayWindow->PostMessage(&m);
		}
		break;
	case opcode::PlayPause:
		{
			Action();
			if (fPlayList->CountFiles() == 0) {
				BMessage m(opcode::PlaySelection);
				MessageReceived(&m);
			} else if (IsPlaying())
				fPlayer->PostMessage(FPInterface::Pause);
			else if (IsPaused())
				fPlayer->PostMessage(FPInterface::Play);
			else {
				if (fPlayList->ContainsFile(fCurrentRef))
					SetCurrent(fCurrentRef, true);
				else {
					entry_ref ref;
					if (fPlayList->GetFirst(ref))
						SetCurrent(ref, true);
				}
			}
		}
		break;
	case opcode::Stop:
		{
			Action();
			fPlayer->PostMessage(FPInterface::Eject);
			if (fPlayList->CountFiles() == 0)
				fMsgWin->ShowText("no file");
			else {
				BMessage m(PlayStatusView::BogusStatus);
				m.AddInt32("status", FPInterface::Empty);
				m.AddRef("ref", &fCurrentRef);
				fPlayWindow->PostMessage(&m);
			}
		}
		break;
	case opcode::Eject:
		Action();
		fMsgWin->ShowText("eject");
		StopPlaying();
		fPlayList->Empty();
		break;
	case opcode::Shuffle:
		Action();
		fPlayList->Shuffle();
		fMsgWin->ShowText("playlist shuffled");
		break;
	case opcode::LaunchedFromAddOn:
		{
			SetPulseRate(kPulseRate);
			Action();
			BMessage m;
			if (msg->FindMessage("msg", &m) == B_OK)
				PostMessage(&m);
		}
		break;
	case opcode::AddSelection:
		{
			Action();
			unsigned num = AddSelection();
			if (num > 0) {
				char buf[100];
				sprintf(buf, (num == 1) ? "added %i file" : "added %i files", int(num));
				fMsgWin->ShowText(buf);
			} else
				fMsgWin->ShowText("no files selected");
		}
		break;
	case opcode::PlaySelection:
		{
			Action();
			StopPlaying();
			fPlayList->Empty();
			unsigned num = AddSelection();
			if (num > 0) {
				char buf[100];
				sprintf(buf, (num == 1) ? "loaded %i file" : "loaded %i files", int(num));
				fMsgWin->SetText(buf);

				entry_ref first;
				if (fPlayList->GetFirst(first))
					SetCurrent(first, true);
				else if (num == 1) {
					fMsgWin->AppendText(", it didn't contain audio");
					fMsgWin->Place();
					fMsgWin->Show();
				} else {
					fMsgWin->AppendText(", none of them contained audio");
					fMsgWin->Place();
					fMsgWin->Show();
				}
			} else
				fMsgWin->ShowText("no file");
		}
		break;

	case FPInterface::StatusNotify:
		{
			Action();
			entry_ref ref;
			if (msg->FindRef("ref", &ref) == B_OK)
				fCurrentRef = ref;
		}
		break;
	case FPInterface::EndOfFileNotify:
		{
			Action();
			entry_ref ref;

			if (msg->FindRef("ref", &ref) != B_OK)
				return;

			entry_ref next;
			if (fPlayList->GetNext(ref, next)) {
				BMessage m(FPInterface::Load);
				m.AddRef("ref", &next);
				m.AddBool("play", true);
				fPlayer->PostMessage(&m);
			} else {
				if (!fPlayList->ContainsFile(ref)) {
					if (fPlayList->GetFirst(next)) {
						BMessage m(FPInterface::Load);
						m.AddRef("ref", &next);
						m.AddBool("play", true);
						fPlayer->PostMessage(&m);
						return;
					}
				}
				fMsgWin->ShowText("end of playlist");
				fPlayer->PostMessage(FPInterface::Eject);
			}
		}
		break;

	default:
		BApplication::MessageReceived(msg);
	}
}

unsigned FancyApp::AddRefs(BMessage *msg, const char *field)
{	ENTRY

	vector<entry_ref> refs;
	refs.push_back();
	for (int i=0; msg->FindRef(field, i, &refs.back()) == B_OK; ++i)
		refs.push_back();
	refs.pop_back();

	return fPlayList->Add(refs, refs.size() == 1);
}

bool FancyApp::IsStopped() const
{	ENTRY

	BMessage reply;
	BMessenger(fPlayer).SendMessage(FPInterface::GetStatus, &reply);
	return reply.what == FPInterface::Empty;
}

bool FancyApp::IsPlaying() const
{	ENTRY

	BMessage reply;
	BMessenger(fPlayer).SendMessage(FPInterface::GetStatus, &reply);
	return reply.what == FPInterface::Playing;
}

bool FancyApp::IsPaused() const
{	ENTRY

	BMessage reply;
	BMessenger(fPlayer).SendMessage(FPInterface::GetStatus, &reply);
	return reply.what == FPInterface::Paused;
}

void FancyApp::SetCurrent(const entry_ref &ref, bool play)
{	ENTRY

	fCurrentRef = ref;
	if (IsPlaying() || play) {
		fPlayer->PostMessage(FPInterface::Eject);
		BMessage m(FPInterface::Load);
		m.AddRef("ref", &ref);
		m.AddBool("play", true);
		fPlayer->PostMessage(&m);
	}
}

void FancyApp::StopPlaying()
{	ENTRY

	fPlayer->PostMessage(FPInterface::Eject);
}

void FancyApp::RefsReceived(BMessage *msg)
{	ENTRY

	StopPlaying();
	fPlayList->Empty();
	unsigned num = AddRefs(msg, "refs");

	if (num == 0)
		fMsgWin->ShowText("no files loaded");
	else if (num == 1)
		fMsgWin->ShowText("1 file loaded");
	else {
		char buf[100];
		sprintf(buf, "%i files loaded", int(num));
		fMsgWin->ShowText(buf);
	}

	if (num != 0) {
		entry_ref first;
		if (fPlayList->GetFirst(first))
			SetCurrent(first, true);
		else if (num == 1)
			fMsgWin->ShowText("it didn't contain audio");
		else
			fMsgWin->ShowText("none of them contained audio");
	}
}

unsigned FancyApp::AddSelection()
{
	int32 wincount;
	BMessage reply;
	BList teams;
	status_t result;
	const bigtime_t HUGE_TIMEOUT = 1000000;

	// Get number of open windows
	BMessenger msgr("application/x-vnd.Be-TRAK");
	BMessage count_msg(B_COUNT_PROPERTIES);
	count_msg.AddSpecifier("Window");
	result = msgr.SendMessage(&count_msg, &reply, HUGE_TIMEOUT, HUGE_TIMEOUT);
	if(result != B_OK)
		return 0;
	result = reply.FindInt32("result", 0, &wincount);
	if(result != B_OK)
		return 0;

	// Check which (if any) window is active
	for(int i=0; i<wincount; i++) {
		BMessage winactive_msg(B_GET_PROPERTY);
		winactive_msg.AddSpecifier("Active");
		winactive_msg.AddSpecifier("Window", i);
		result = msgr.SendMessage(&winactive_msg, &reply, HUGE_TIMEOUT, HUGE_TIMEOUT);
		if(result == B_OK) {
			bool active;
			result = reply.FindBool("result", 0, &active);
			if(result == B_OK && active == true) {
				// Found the active window, get the selection
				BMessage sel_msg(B_GET_PROPERTY);
				sel_msg.AddSpecifier("Selection");
				sel_msg.AddSpecifier("View", "PoseView");
				sel_msg.AddSpecifier("Window", i);
				result = msgr.SendMessage(&sel_msg, &reply, HUGE_TIMEOUT, HUGE_TIMEOUT);
				if(result != B_OK)
					break;
				return AddRefs(&reply, "result");
					// Add selection to the playlist
			}
		}
	}
	return 0;
}

void FancyApp::Action()
{	ENTRY

	fLastAction = system_time();
}

int main(int argc, char **argv)
{	ENTRY

	srand48(time(NULL));

	FancyApp app;

	app.Run();
}
