#include "FancyDebug.h"

#include <Directory.h>
#include <Mime.h>
#include <Path.h>
#include <Query.h>
#include <StorageDefs.h>
#include <String.h>
#include <VolumeRoster.h>

#include <algorithm>
#include <fstream>

#include <kernel/fs_attr.h>

#include "FancyPlayList.h"
#include "FancyPlayer.h"

static const char *kPlayListMimes[] = {
	"text/x-playlist",
	"text/x-soundplay-playlist",
	"text/x-cl-amp-playlist",
	0
};

static const char *kQueryMime = "application/x-vnd.Be-query";
static const char *kAttrQueryVolume = "_trk/qryvol1";

// from OpenTracker:
static status_t MatchArchivedVolume(BVolume *result, const BMessage *message, int32 index)
{	ENTRY

	time_t created;
	off_t capacity;
	
	if (message->FindInt32("creationDate", index, &created) != B_OK
		|| message->FindInt64("capacity", index, &capacity) != B_OK)
		return B_ERROR;

	BVolumeRoster roster;
	BVolume volume;

	roster.Rewind();
	while (roster.GetNextVolume(&volume) == B_OK)
		if (volume.IsPersistent() && volume.KnowsQuery()) {
			BDirectory root;
			volume.GetRootDirectory(&root);
			time_t cmpCreated;
			root.GetCreationTime(&cmpCreated);
			if (created == cmpCreated && capacity == volume.Capacity()) {
				*result = volume;
				return B_OK;
			}
		}

	return B_DEV_BAD_DRIVE_NUM;
}

static unsigned AddQuery(const BString &qrystr, const BVolume &volume, FancyPlayList::entry_vec &audioFiles)
{	ENTRY

	BQuery q;
	q.SetPredicate(qrystr.String());
	q.SetVolume(&volume);
	if (q.Fetch() != B_OK)
		return 0;

	audioFiles.push_back();
	unsigned num = 0;
	while (q.GetNextRef(&audioFiles.back()) == B_OK) {
		audioFiles.push_back();
		++num;
	}
	audioFiles.pop_back();

	return num;
}

static unsigned AddQuery(const entry_ref &ref, FancyPlayList::entry_vec &audioFiles)
{	ENTRY

	BEntry entry(&ref);
	if (entry.InitCheck() != B_OK)
		return 0;

	BNode node(&ref);
	if (node.InitCheck() != B_OK)
		return false;

	BString mime;
	if (node.ReadAttrString("BEOS:TYPE", &mime) != B_OK)
		return false;

	BString qrystr;
	if (node.ReadAttrString("_trk/qrystr", &qrystr) != B_OK)
		return false;

	bool searchAllVolumes = true;
	status_t result = B_OK;
	unsigned num = 0;

	// from OpenTracker:

	// get volumes to perform query on
	attr_info info;
	if (node.GetAttrInfo(kAttrQueryVolume, &info) == B_OK) {
		BString buffer;
		if (node.ReadAttr(kAttrQueryVolume, B_MESSAGE_TYPE, 0,
			buffer.LockBuffer((int32)info.size), (size_t)info.size) == info.size) {
			buffer.UnlockBuffer();
			BMessage message;
			if (message.Unflatten(buffer.String()) == B_OK) {
				for (int32 index = 0; ;index++) {
					ASSERT(index < 100);
					BVolume volume;
						// match a volume with the info embedded in the message
					result = MatchArchivedVolume(&volume, &message, index);
					if (result == B_OK) {
						// send the querry off on this volume
						num += AddQuery(qrystr, volume, audioFiles);

						searchAllVolumes = false;
					} else if (result != B_DEV_BAD_DRIVE_NUM)
						// if B_DEV_BAD_DRIVE_NUM, the volume just isn't mounted this
						// time around, keep looking for more
						// if other error, bail
						break;
				}
			}
		}
	}

	if (searchAllVolumes) {
		// no specific volumes embedded in query, search everything
		BVolumeRoster roster;
		BVolume volume;
	
		roster.Rewind();
		while (roster.GetNextVolume(&volume) == B_OK)
			if (volume.IsPersistent() && volume.KnowsQuery())
				num += AddQuery(qrystr, volume, audioFiles);
	}

	return num;
}

static unsigned AddPlayList(const entry_ref &ref, FancyPlayList::entry_vec &audioFiles)
{	ENTRY

	BEntry entry(&ref);
	if (entry.InitCheck() != B_OK)
		return 0;

	BNode node(&ref);
	if (node.InitCheck() != B_OK)
		return false;
	BString mime;
	if (node.ReadAttrString("BEOS:TYPE", &mime) != B_OK)
		return false;
	if (!strcmp(mime.String(), kQueryMime))
		return AddQuery(ref, audioFiles);

	BEntry p;
	if (entry.GetParent(&p) != B_OK)
		return 0;
	BPath parentPath;
	if (p.GetPath(&parentPath) != B_OK)
		return 0;

	chdir(parentPath.Path());
		// so that relative paths in the playlist works

	BPath path;
	if (entry.GetPath(&path) != B_OK)
		return 0;

	ifstream file(path.Path());
	unsigned num = 0;
	if (file && file.is_open()) {
		char buf[B_PATH_NAME_LENGTH];

		while(file.getline(buf, B_PATH_NAME_LENGTH)) {
			BEntry e(buf);
			if (e.InitCheck() != B_OK)
				continue;
			if (!e.Exists())
				continue;
			entry_ref r;
			if (e.GetRef(&r) == B_OK) {
				audioFiles.push_back(r);
				++num;
			}
		}
	}
	return num;
}

static bool IsPlayList(const entry_ref &ref)
{	ENTRY

	BNode node(&ref);
	if (node.InitCheck() != B_OK)
		return false;

	BString mime;
	if (node.ReadAttrString("BEOS:TYPE", &mime) != B_OK)
		return false;

	for (const char **p = kPlayListMimes; *p; ++p) {
		if (!strcmp(mime.String(), *p))
			return true;
	}
	if (!strcmp(mime.String(), kQueryMime))
		return true;
	return false;
}

static unsigned Analyze(const entry_ref &ref, FancyPlayList::entry_vec &audioFiles)
{	ENTRY

	if (IsAudioFile(ref)) {
		audioFiles.push_back(ref);
		return 1;
	} else if (IsPlayList(ref))
		return AddPlayList(ref, audioFiles);
	else
		return 0;
}

static unsigned Analyze(const FancyPlayList::entry_vec &refs, FancyPlayList::entry_vec &audioFiles)
{	ENTRY

	bool num = 0;
	for (FancyPlayList::entry_vec::const_iterator it = refs.begin(); it != refs.end(); ++it)
		num += Analyze(*it, audioFiles);
	return num;
}

static unsigned Clean(FancyPlayList::entry_vec &refs, FancyPlayList::entry_vec &removed)
{	ENTRY

	unsigned num = 0;
	for (unsigned i=0; i<refs.size(); ) {
		entry_ref &ref = refs[i];
		if (IsAudioFile(ref) || IsPlayList(ref))
			++i;
		else {
			refs.erase(refs.begin() + i);
			++num;
		}
	}
	return num;
}

unsigned FancyPlayList::Clean(entry_vec *removed)
{	ENTRY

	entry_vec _res;
	return ::Clean(fList, removed ? *removed : _res);
}

unsigned FancyPlayList::Add(const entry_ref &ref, bool analyze, entry_vec *added)
{	ENTRY

	if (analyze) {
		entry_vec _added;
		Analyze(ref, _added);
		fList.insert(fList.end(), _added.begin(), _added.end());
		if (added)
			added->insert(added->begin(), _added.begin(), _added.end());
		return _added.size();
	} else {
		fList.push_back(ref);
		if (added)
			added->push_back(ref);
		return 1;
	}
}

unsigned FancyPlayList::Add(const entry_vec &refs, bool analyze, entry_vec *added)
{	ENTRY

	if (analyze) {
		entry_vec _added;
		Analyze(refs, _added);
		fList.insert(fList.end(), _added.begin(), _added.end());
		if (added)
			added->insert(added->begin(), _added.begin(), _added.end());
		return _added.size();
	} else {
		fList.insert(fList.end(), refs.begin(), refs.end());
		if (added)
			added->insert(added->end(), refs.begin(), refs.end());
		return refs.size();
	}
}

bool FancyPlayList::ContainsFile(const entry_ref &ref) const
{	ENTRY

	return find(fList.begin(), fList.end(), ref) != fList.end();
}

unsigned FancyPlayList::Remove(const entry_ref &ref, entry_vec *removed)
{	ENTRY

	entry_vec::iterator it = find(fList.begin(), fList.end(), ref);
	if (it == fList.end())
		return 0;
	else {
		fList.erase(it);
		if (removed)
			removed->push_back(ref);
		return 1;
	}
}

unsigned FancyPlayList::Remove(const entry_vec &refs, entry_vec *removed)
{	ENTRY

	unsigned num = 0;
	for (entry_vec::const_iterator it = refs.begin(); it != refs.end(); ++it)
		num += Remove(*it, removed);
	return num;
}

unsigned FancyPlayList::Empty(entry_vec *removed)
{	ENTRY

	if (removed)
		removed->insert(removed->end(), fList.begin(), fList.end());

	unsigned num = fList.size();
	fList.clear();
	return num;
}

void FancyPlayList::Shuffle()
{	ENTRY

	random_shuffle(fList.begin(), fList.end());
}

bool FancyPlayList::GetFirst(entry_ref &result, entry_vec *added, entry_vec *removed)
{	ENTRY

	if (fList.size() == 0)
		return false;
	if (IsAudioFile(fList.front())) {
		result = fList.front();
		return true;
	}

	entry_ref current = fList.front();
	entry_vec newRefs;
	::Analyze(fList.front(), newRefs);
	fList.insert(fList.begin(), newRefs.begin(), newRefs.end());
	Remove(current, removed);
	if (added) added->insert(added->end(), newRefs.begin(), newRefs.end());

	return GetFirst(result, added, removed);
}

bool FancyPlayList::GetLast(entry_ref &result, entry_vec *added, entry_vec *removed)
{	ENTRY

	if (fList.size() == 0)
		return false;
	if (IsAudioFile(fList.back())) {
		result = fList.back();
		return true;
	}

	entry_ref current = fList.back();
	entry_vec newRefs;
	::Analyze(fList.back(), newRefs);
	fList.insert(fList.end(), newRefs.begin(), newRefs.end());
	Remove(current, removed);
	if (added) added->insert(added->end(), newRefs.begin(), newRefs.end());

	return GetLast(result, added, removed);
}

bool FancyPlayList::GetNext(const entry_ref &from, entry_ref &result, entry_vec *added, entry_vec *removed)
{	ENTRY

	entry_vec::iterator it = find(fList.begin(), fList.end(), from);
	if (it == fList.end())
		return false;
	++it;
	if (it == fList.end())
		return false;

	if (IsAudioFile(*it)) {
		result = *it;
		return true;
	}

	entry_ref current = *it;
	entry_vec newRefs;
	::Analyze(current, newRefs);
	fList.insert(it, newRefs.begin(), newRefs.end());
	Remove(current, removed);
	if (added) added->insert(added->end(), newRefs.begin(), newRefs.end());

	return GetNext(from, result, added, removed);
}

bool FancyPlayList::GetPrevious(const entry_ref &from, entry_ref &result, entry_vec *added, entry_vec *removed)
{	ENTRY

	entry_vec::reverse_iterator it = find(fList.rbegin(), fList.rend(), from);
	if (it == fList.rend())
		return false;
	++it;
	if (it == fList.rend())
		return false;

	if (IsAudioFile(*it)) {
		result = *it;
		return true;
	}

	entry_ref current = *it;
	entry_vec newRefs;
	::Analyze(current, newRefs);
	fList.insert(it.base(), newRefs.begin(), newRefs.end());
	Remove(current, removed);
	if (added) added->insert(added->end(), newRefs.begin(), newRefs.end());

	return GetPrevious(from, result, added, removed);
}
