#include "FancyDebug.h"

#include <MediaFile.h>
#include <MediaTrack.h>
#include <MessageQueue.h>
#include <Roster.h>
#include <SoundPlayer.h>
#include <String.h>

#include <iostream>
#include <algorithm>

#include "FancyPlayer.h"
#include "FancyPlayerInterface.h"

FancyPlayer::FileData::FileData(const entry_ref &_ref)
	:	isValid(false),
		ref(_ref),
		file(0),
		track(0)
{	ENTRY

	file = new BMediaFile(&ref);
	if (file->InitCheck() != B_OK)
		return;

	int32 numTracks = file->CountTracks();
	for (int32 i = 0; i < numTracks; ++i) {
		BMediaTrack *_track = file->TrackAt(i);

		if (_track == NULL)
			return;
		else {
			if (_track->EncodedFormat(&format) == B_OK) {
				switch (format.type) {
				case B_MEDIA_ENCODED_AUDIO:
					if (_track->DecodedFormat(&format) != B_OK)
						break;
					// fallthru
				case B_MEDIA_RAW_AUDIO:
					track = _track;
					isValid = true;
					MESSAGE("file contains audio")
					return;

				default:
					break;
				}
			}
			file->ReleaseTrack(_track);
		}
	}

}

FancyPlayer::FileData::~FileData()
{	DENTRY
	delete file;
}

FancyPlayer::StatusMonitor::StatusMonitor(FancyPlayer *_which)
	:	which(_which)
{	ENTRY
}

void FancyPlayer::StatusMonitor::Mark()
{	ENTRY

	status = which->GetStatus();
	if (which->HasFile())
		ref = which->fFileData->ref;
}

uint32 FancyPlayer::StatusMonitor::Changed() const
{	ENTRY
	using namespace FPInterface;

	entry_ref compRef;
	if (which->HasFile())
		compRef = which->fFileData->ref;

	uint32 flags = 0;
	if (status != which->GetStatus() || compRef != ref)
		flags |= StatusBit;

	MESSAGE("flags: " << flags)

	return flags;
}

FancyPlayer::AutoNotify::AutoNotify(FancyPlayer *_which)
	:	which(_which),
		mon(_which)
{	ENTRY

	which->fAutoNotifyStack++;
	MESSAGE("stack depth: " << which->fAutoNotifyStack)

	if (which->fAutoNotifyStack == 1)
		mon.Mark();
}

FancyPlayer::AutoNotify::~AutoNotify()
{	DENTRY
	using namespace FPInterface;

	which->fAutoNotifyStack--;
	MESSAGE("stack depth: " << which->fAutoNotifyStack)

	if (which->fAutoNotifyStack != 0)
		return;

	uint32 flags = mon.Changed();

	if (flags)
		which->Notify(flags);
}

FancyPlayer::FancyPlayer(const char *name)
	:	BLooper(name),
		fAutoNotifyStack(0),
		fPaused(false),
		fFileData(0),
		fPlayer(0),
		fAudioLock("fAudioLock")
{	ENTRY
}

FancyPlayer::~FancyPlayer()
{	DENTRY

	Eject();
	Notify(QuitBit);
}

FPInterface::Status FancyPlayer::GetStatus() const
{	ENTRY
	using namespace FPInterface;

	if (HasFile() && fPlayer) {
		if (fPaused) {
			MESSAGE("paused");
			return Paused;
		} else {
			MESSAGE("playing");
			return Playing;
		}
	} else {
		MESSAGE("empty");
		return Empty;
	}
}

status_t FancyPlayer::Load(const entry_ref &ref)
{	ENTRY

	launch_media_server();

	FileData *fileData = new FileData(ref);
	if (!fileData->isValid) {
		delete fileData;
		MESSAGE("isValid == false")
		return B_MEDIA_BAD_FORMAT;
	}
	MESSAGE("isValid == true")

	AutoNotify anObj(this);
	Eject();

	fAudioLock.Lock();

	fFileData = fileData;

	media_format format;
	fPerfTime = -1;
	fTrackTime = 0;
	fFinalTime = fFileData->track->Duration() - 1;

	fFileData->track->DecodedFormat(&format);
	switch (format.u.raw_audio.format) {
	case media_raw_audio_format::B_AUDIO_UCHAR:
		fDefaultData = 0x80;
		fFrameSize = 1;
		break;
	case media_raw_audio_format::B_AUDIO_SHORT:
		fDefaultData = 0;
		fFrameSize = 2;
		break;
	case media_raw_audio_format::B_AUDIO_INT:
		fDefaultData = 0;
		fFrameSize = 4;
		break;
	case media_raw_audio_format::B_AUDIO_FLOAT:
		fDefaultData = 0;
		fFrameSize = 4;
		break;
	default:
		return B_MEDIA_BAD_FORMAT;
	}

	fFrameSize *= format.u.raw_audio.channel_count;
	fBufferSize = format.u.raw_audio.buffer_size;
	fFrameRate = format.u.raw_audio.frame_rate;

	fPlayer = new BSoundPlayer(&format.u.raw_audio, ref.name, FillBuffer);

	status_t err = fPlayer->InitCheck();
	if (err != B_OK) {
		delete fPlayer;
		fPlayer = 0;
		fPaused = false;
	} else {
		fPlayer->SetCookie(this);
		fPaused = true;
	}

	MESSAGE(strerror(err))

	fAudioLock.Unlock();

	return err;
}

void FancyPlayer::MessageReceived(BMessage *msg)
{	ENTRY

	switch (msg->what) {
	case FPInterface::GetTime:
		if (HasFile()) {
			BMessage m(B_OK);
			fAudioLock.Lock();
			status_t err = m.AddInt64("time", fTrackTime);
			fAudioLock.Unlock();
			if (err == B_OK)
				msg->SendReply(&m);
			else
				msg->SendReply(err);
		} else
			msg->SendReply(B_NO_INIT);
		break;
	case FPInterface::Seek:
		{
			bigtime_t time;
			status_t err = msg->FindInt64("time", &time);

			bool sloppy;
			if (err == B_OK)
				if (msg->FindBool("sloppy", &sloppy) != B_OK)
					sloppy = true;

			if (err == B_OK)
				err = Seek(time, sloppy);

			BMessage m(B_OK);
			if (err == B_OK) {
				err = m.AddInt64("time", time);
			}

			if (err == B_OK)
				msg->SendReply(&m);
			else
				msg->SendReply(err);
		}
		break;
	case EndOfTrackReached:
		{
			fPlayer->Stop();
			fPaused = true;

			// we already know we reached the end, don't nag.
			// All EndOfTrackReached messages will be in the queue
			// now because of the blocking call to BSoundPlayer::Stop()
			BMessageQueue *queue = MessageQueue();
			queue->Lock();
			BMessage *m;
			while ((m = queue->FindMessage(EndOfTrackReached, 0)))
				queue->RemoveMessage(m);
			queue->Unlock();

			Notify(StatusBit | EndOfTrackBit);
		}
		break;
	case FPInterface::Eject:
		{
			StatusMonitor mon(this);
			mon.Mark();
			BMessage m(Eject());
			status_t err = m.AddInt32("change", mon.Changed());
			if (err == B_OK)
				msg->SendReply(&m);
			else
				msg->SendReply(err);
		}
		break;
	case FPInterface::Pause:
		{
			StatusMonitor mon(this);
			mon.Mark();
			BMessage m(Pause());
			status_t err = m.AddInt32("change", mon.Changed());
			if (err == B_OK)
				msg->SendReply(&m);
			else
				msg->SendReply(err);
		}
		break;
	case FPInterface::Play:
		{
			StatusMonitor mon(this);
			mon.Mark();
			BMessage m(Play());
			status_t err = m.AddInt32("change", mon.Changed());
			if (err == B_OK)
				msg->SendReply(&m);
			else
				msg->SendReply(err);
		}
		break;
	case FPInterface::Load:
		{
			AutoNotify anObj(this);
			entry_ref ref;
			status_t err = msg->FindRef("ref", &ref);
			if (err != B_OK)
				msg->SendReply(err);
			else {
				msg->SendReply(Load(ref));
				bool play;
				if (msg->FindBool("play", &play) == B_OK)
					if (play)
						Play();
			}
		}
		break;
	case FPInterface::GetStatus:
		msg->SendReply(GetStatus());
		break;
	case FPInterface::GetFileInfo:
		if (HasFile()) {
			BMessage m(B_OK);

			fAudioLock.Lock();
			status_t err = m.AddInt64("length", fFinalTime+1);
			fAudioLock.Unlock();
			if (err != B_OK)
				msg->SendReply(err);
			else {
				err = m.AddRef("ref", &fFileData->ref);
				if (err != B_OK)
					msg->SendReply(err);
				else
					msg->SendReply(&m);
			}
		} else
			msg->SendReply(B_FILE_ERROR);
		break;
	case FPInterface::RegisterCallback:
		{
			BMessenger msngr;
			status_t err = msg->FindMessenger("messenger", &msngr);
			if (err == B_OK)
				msg->SendReply(RegisterCallback(msngr));
			else
				msg->SendReply(err);
		}
		break;
	case FPInterface::RemoveCallback:
		{
			BMessenger msngr;
			status_t err = msg->FindMessenger("messenger", &msngr);
			if (err == B_OK)
				msg->SendReply(RemoveCallback(msngr));
			else
				msg->SendReply(err);
		}
		break;
	case FPInterface::GetCallbacks:
		{
			Notify(0);
				// make sure no invalid messengers are hanging around

			BMessage m(B_OK);
			for (notifier_list_t::iterator it = fNotifierList.begin(); it != fNotifierList.end(); ++it) {
				status_t err = m.AddMessenger("messengers", *it);
				if (err != B_OK) {
					msg->SendReply(err);
					break;
				}
			}
			msg->SendReply(&m);
		}
		break;

	default:
		BLooper::MessageReceived(msg);
	}
}

void FancyPlayer::FillBuffer(void *cookie, void *buffer, size_t bufferSize, const media_raw_audio_format &format)
{
	static_cast<FancyPlayer *>(cookie)->FillBuffer(buffer, bufferSize, format);
}

void FancyPlayer::FillBuffer(void *buffer, size_t bufferSize, const media_raw_audio_format &format)
{
	DEBUG_ONLY(
	static bool __fd_called = false;
	if (!__fd_called) {
		MESSAGE("FillBuffer called, this is the only notification")
		__fd_called = true;
	}
	)

	fAudioLock.Lock();

	int64			frameCount;
	bool			updateTrackTime = true;

	status_t err = fFileData->track->ReadFrames(buffer, &frameCount);
	if (err != B_OK || frameCount < 0) {
		memset(buffer, fDefaultData, fBufferSize);
		updateTrackTime = false;

		fFinalTime = fFileData->track->CurrentTime();
		PostMessage(EndOfTrackReached);
	} else {
		uint32 filled = fFrameSize * frameCount;
		if (filled < fBufferSize)
			memset(static_cast<char *>(buffer)+filled, fDefaultData, fBufferSize - filled);
	}

	fPerfTime = fPlayer->PerformanceTime();
	if (updateTrackTime)
		fTrackTime = fFileData->track->CurrentTime();
	else
		fTrackTime += bigtime_t(1.0e6*float(fBufferSize)/float(fFrameSize * fFrameRate));

	fAudioLock.Unlock();
}

bool IsAudioFile(const entry_ref &ref)
{	ENTRY

	FancyPlayer::FileData fileData(ref);
	return fileData.isValid;
}

static void NotifySingle(BMessage *stat, BMessage *eot, BMessage *quit, BMessenger *msngr)
{	ENTRY

	if (stat) msngr->SendMessage(stat);
	if (eot) msngr->SendMessage(eot);
	if (quit) msngr->SendMessage(quit);
}

void FancyPlayer::Notify(uint32 flags, BMessenger *singleMessenger)
{	ENTRY
	using namespace FPInterface;

	BMessage statMsg(StatusNotify);
	BMessage quitMsg(QuitNotify);
	BMessage eotMsg(EndOfFileNotify);

	if (flags & StatusBit) {
		MESSAGE("status")
		statMsg.AddInt32("status", GetStatus());
		statMsg.AddBool("ended", flags & EndOfTrackBit);

		if (HasFile()) {
			statMsg.AddRef("ref", &fFileData->ref);
			statMsg.AddInt64("length", fFinalTime+1);
		}
	}

	if (flags & EndOfTrackBit) {
		MESSAGE("end of track")
		eotMsg.AddInt64("length", fFinalTime+1);
		if (HasFile())
			eotMsg.AddRef("ref", &fFileData->ref);
	}

	DEBUG_ONLY(
	if (flags & QuitBit) {
		MESSAGE("quit")
	}	
	)

	if (singleMessenger)
		NotifySingle((StatusBit & flags) ? &statMsg : 0, (EndOfTrackBit & flags) ? &eotMsg : 0, (QuitBit & flags) ? &quitMsg : 0, singleMessenger);
	else {
		for (unsigned i = 0; i<fNotifierList.size(); )
			if (fNotifierList[i].IsValid()) {
				NotifySingle((StatusBit & flags) ? &statMsg : 0, (EndOfTrackBit & flags) ? &eotMsg : 0, (QuitBit & flags) ? &quitMsg : 0, &fNotifierList[i]);
				++i;
			} else
				fNotifierList.erase(fNotifierList.begin() + i);
	}
}

status_t FancyPlayer::Pause()
{	ENTRY

	if (GetStatus() != FPInterface::Playing)
		return B_ERROR;

	AutoNotify anObj(this);

	fPaused = true;
	fPlayer->SetHasData(false);
	fPlayer->Stop();

	return B_OK;
}

status_t FancyPlayer::Play()
{	ENTRY
	using namespace FPInterface;

	Status status = GetStatus();
	if ((status != Paused) || !fPlayer)
		return B_ERROR;

	AutoNotify anObj(this);

	status_t err = fPlayer->Start();
	if (err == B_OK) {
		fPlayer->SetHasData(true);
		fPaused = false;
	}

	MESSAGE(strerror(err))

	return err;
}

status_t FancyPlayer::Eject()
{	ENTRY

	if (GetStatus() == FPInterface::Empty)
		return B_ERROR;

	AutoNotify anObj(this);

	Pause();

	delete fFileData;
	delete fPlayer;
	fFileData = 0;
	fPlayer = 0;

	return B_OK;
}

status_t FancyPlayer::Seek(bigtime_t &time, bool sloppy)
{	ENTRY

	if (GetStatus() == FPInterface::Empty)
		return B_ERROR;

	if (!sloppy)
		return B_UNSUPPORTED;

	fAudioLock.Lock();
	status_t err = fFileData->track->SeekToTime(&time);
	fAudioLock.Unlock();

	MESSAGE(strerror(err))

	return err;
}

status_t FancyPlayer::RegisterCallback(BMessenger &msngr)
{	ENTRY

	notifier_list_t::iterator it = find(fNotifierList.begin(), fNotifierList.end(), msngr);

	if (it == fNotifierList.end())
		fNotifierList.push_back(msngr);

	Notify(StatusBit, &msngr);

	return B_OK;	
}

status_t FancyPlayer::RemoveCallback(const BMessenger &msngr)
{	ENTRY

	notifier_list_t::iterator it = find(fNotifierList.begin(), fNotifierList.end(), msngr);
	if (it == fNotifierList.end())
		return B_ERROR;

	fNotifierList.erase(it);

	return B_OK;	
}
