#include "FancyDebug.h"

#include <Message.h>
#include <Screen.h>

#include <algorithm>

#include "TimedWindow.h"

BLocker TimedWindow::fAllTimedLock(true);
vector<TimedWindow *> TimedWindow::fAllTimed;
bigtime_t TimedWindow::fHideTime = 5000000;
thread_id TimedWindow::fWaiter = 0;

TimedWindow::TimedWindow(BRect frame, const char *title, window_type type, uint32 flags, uint32 workspaces)
	:	BWindow(frame, title, type, flags, workspaces)
{	ENTRY

	ConstructorCommon();
}

TimedWindow::TimedWindow(BRect frame, const char *title, window_look look, window_feel feel, uint32 flags, uint32 workspaces)
	:	BWindow(frame, title, look, feel, flags, workspaces)
{	ENTRY

	ConstructorCommon();
}

TimedWindow::~TimedWindow()
{	DENTRY

	fAllTimedLock.Lock();

	vector<TimedWindow *>::iterator it = find(fAllTimed.begin(), fAllTimed.end(), this);
	if (it != fAllTimed.end())
		fAllTimed.erase(it);

	suspend_thread(fWaiter);
	resume_thread(fWaiter);

	unsigned num = fAllTimed.size();

	fAllTimedLock.Unlock();

	if (num == 0) {
		int32 ret;
		MESSAGE("waiting for thread")
		wait_for_thread(fWaiter, &ret);
		MESSAGE("it returned")
	}
}

void TimedWindow::MessageReceived(BMessage *msg)
{	ENTRY

	switch (msg->what) {
	case ShowWindow:
		Show();
		break;
	case HideWindow:
		Hide();
		break;

	default:
		BWindow::MessageReceived(msg);
	}
}

void TimedWindow::Place()
{	ENTRY

	const BRect wFrame = Frame();
	const BRect sFrame = BScreen().Frame();
	MoveTo((sFrame.right - wFrame.Width()) / 2.0f, sFrame.bottom - wFrame.Height());
}

void TimedWindow::WorkspaceActivated(int32 workspace, bool active)
{	ENTRY

	Place();
}

void TimedWindow::ConstructorCommon()
{	ENTRY

	Lock();
	MoveBy(0, -10000);
	Show();
	Hide();
	MoveBy(0, 10000);
	Unlock();

	fAllTimedLock.Lock();

	fAllTimed.push_back(this);

	fAutoHiding = true;
	fWhenToHide = -1;

	if (fWaiter == 0) {
		fWaiter = spawn_thread(WaiterFunc, "hide 'n' sleep", B_NORMAL_PRIORITY, 0);
		if (fWaiter <= 0)
			FatalError("Failed to spawn hide 'n' sleep thread");
		resume_thread(fWaiter);
	}

	fAllTimedLock.Unlock();
}

int32 TimedWindow::WaiterFunc()
{	ENTRY

	bigtime_t nextTime = 0;

	while (true) {
		fAllTimedLock.Lock();
		if (fAllTimed.size() == 0) {
			MESSAGE("*goodbye")
			fAllTimedLock.Unlock();
			return 0;
		}

		bigtime_t systime = system_time();

		MESSAGE("*hiding windows")
		nextTime = -1;
		for (vector<TimedWindow *>::iterator it = fAllTimed.begin(); it != fAllTimed.end(); ++it) {
			TimedWindow &cur = **it;
			cur.Lock();
			if (cur.fAutoHiding) {
				if (cur.fWhenToHide < systime) {
					if (!cur.IsHidden())
						cur.Hide();
					cur.fWhenToHide = -1;
				} else
					nextTime = cur.fWhenToHide;
			} else
				cur.fWhenToHide = -1;
			cur.Unlock();
		}

		fAllTimedLock.Unlock();

		if (nextTime == -1) {
			MESSAGE("*taking a nap")
			suspend_thread(find_thread(0));
		} else {
			MESSAGE("*taking a nap for " << nextTime - systime << "us")
			DEBUG_ONLY( status_t err = ) snooze_until(nextTime, B_SYSTEM_TIMEBASE);
			MESSAGE(strerror(err))
		}
		MESSAGE("*good morning")
	}

	return 0;
}

int32 TimedWindow::WaiterFunc(void *data)
{
	return static_cast<TimedWindow *>(data)->WaiterFunc();
}

void TimedWindow::SetAutoHiding(bool ah)
{	ENTRY

	fAutoHiding = ah;
}

void TimedWindow::SetHideTime(bigtime_t t)
{	ENTRY

	fHideTime = t;
}

void TimedWindow::Show()
{	ENTRY

	Lock();
	fAllTimedLock.Lock();

	for (vector<TimedWindow *>::iterator it = fAllTimed.begin(); it != fAllTimed.end(); ++it)
		if (*it != this)
			(*it)->Hide();

	if (IsHidden())
		BWindow::Show();
	HideIn(fHideTime);
	fAllTimedLock.Unlock();
	Unlock();
}

void TimedWindow::Hide()
{	ENTRY

	Lock();
	fAllTimedLock.Lock();
	fWhenToHide = -1;
	if (!IsHidden())
		BWindow::Hide();
	fAllTimedLock.Unlock();
	Unlock();
}

void TimedWindow::HideIn(bigtime_t t)
{	ENTRY

	HideAt(system_time() + t);
}

void TimedWindow::HideAt(bigtime_t t)
{	ENTRY

	fAllTimedLock.Lock();

	fWhenToHide = t;
	suspend_thread(fWaiter);
	resume_thread(fWaiter);

	fAllTimedLock.Unlock();
}
