#include "FancyDebug.h"

#include <Alert.h>
#include <MediaRoster.h>
#include <Messenger.h>
#include <ParameterWeb.h>
#include <String.h>

#include "Opcodes.h"
#include "VolumeView.h"

VolumeView::VolumeView(const BRect &frame, const char *name)
	:	BView(frame, name, B_FOLLOW_ALL, B_WILL_DRAW)
{	ENTRY

	SetViewColor(B_TRANSPARENT_32_BIT);
}

void VolumeView::AttachedToWindow()
{	ENTRY

	BMediaRoster *roster = BMediaRoster::Roster();
	media_node sysout;
	if (roster->GetAudioOutput(&sysout) != B_OK)
		FatalError("Unable to get audio output");

	BParameterWeb *web;
	if (roster->GetParameterWebFor(sysout, &web) != B_OK)
		FatalError("Unable to get parameter web");

	int mixParamNum = -1;
	for (int i=0; i<web->CountParameters(); i++) {
		if (!strcmp(web->ParameterAt(i)->Name(), "Mixer")) {
			// I'm sure there is a better way to find the system Mixer,
			// but this will do for now.
			mixParamNum = i;
			break;
		}
	}

	if (mixParamNum == -1)
		FatalError("Did not find system mixer");

	mutePar = dynamic_cast<BDiscreteParameter *>(web->ParameterAt(mixParamNum + 1));
	if (!mutePar)
		FatalError("Unable to get mute parameter");

	volumePar = dynamic_cast<BContinuousParameter *>(web->ParameterAt(mixParamNum + 2));
	if (!volumePar)
		FatalError("Unable to get volume parameter");

	if (mutePar->ValueType() != B_INT32_TYPE)
		FatalError("Unsupported value type for mute parameter");

	if (volumePar->ValueType() != B_FLOAT_TYPE)
		FatalError("Unsupported value type for volume parameter");

	if (volumePar->CountChannels() != 2 || mutePar->CountChannels() != 1)
		FatalError("Unsupported number of channels");

	volumePar->GetResponse(&gainResponse, &gainFactor, &gainOffset);
	aMinGain = volumePar->MinValue();
	aMaxGain = volumePar->MaxValue();
	minGain = TransformGain(aMinGain);
	maxGain = TransformGain(aMaxGain);

	status_t err;
	if ((err = roster->StartWatching(BMessenger(this, Window()), B_MEDIA_PARAMETER_CHANGED)) != B_OK) {
		BString txt = "Failed to watch parameter changes: ";
		txt += strerror(err);
		FatalError(txt.String());
	}

	UpdateVars();
}

VolumeView::~VolumeView()
{	DENTRY

	BMediaRoster::Roster()->StopWatching(BMessenger(this, Window()), B_MEDIA_PARAMETER_CHANGED);
}

void VolumeView::MessageReceived(BMessage *msg)
{	ENTRY

	switch (msg->what) {
	case B_MEDIA_PARAMETER_CHANGED:
		UpdateVars();
		Invalidate();
		break;
	case opcode::VolumeUp:
		UpdateVars();
		actualGain[0] += (aMaxGain - aMinGain) / 50.0;
		actualGain[1] += (aMaxGain - aMinGain) / 50.0;
		if (actualGain[0] > aMaxGain) actualGain[0] = aMaxGain;
		if (actualGain[1] > aMaxGain) actualGain[1] = aMaxGain;
		volumePar->SetValue(actualGain, sizeof(float)*2, 0);
		UpdateVars();
		Invalidate();
		Window()->Show();
		break;
	case opcode::VolumeDown:
		UpdateVars();
		actualGain[0] -= (aMaxGain - aMinGain) / 50.0;
		actualGain[1] -= (aMaxGain - aMinGain) / 50.0;
		if (actualGain[0] < aMinGain) actualGain[0] = aMinGain;
		if (actualGain[1] < aMinGain) actualGain[1] = aMinGain;
		volumePar->SetValue(actualGain, sizeof(float)*2, 0);
		UpdateVars();
		Invalidate();
		Window()->Show();
		break;
	case opcode::Mute:
		UpdateVars();
		muted = !muted;
		mutePar->SetValue(&muted, sizeof(int32), 0);
		UpdateVars();
		Invalidate();
		Window()->Show();
		break;
	default:
		BView::MessageReceived(msg);
	}
}

void VolumeView::Draw(BRect updateRect)
{	ENTRY

	if (muted)
		SetHighColor(255,0,0);
	else
		SetHighColor(0,255,0);
	SetLowColor(0,0,0);

	const BRect bounds = Bounds();
	const int barWidth = 4;
	const int barSpace = 2;
	const int numBars = (bounds.IntegerWidth()+barSpace)/(barWidth+barSpace);
	const float _gain = (gain[0] + gain[1])/2.0f;
	const int volume = int(((_gain-minGain)/(maxGain-minGain))*numBars);
	const int maxHeight = bounds.IntegerHeight();

	int x=0;
	for (int i=0; i<=volume; i++) {
		const int height = (i*maxHeight)/numBars;
		const int y = maxHeight - height;
		FillRect(BRect(x,0,x+barWidth-1,y-1), B_SOLID_LOW);
		FillRect(BRect(x,y,x+barWidth-1,maxHeight), B_SOLID_HIGH);

		x += barWidth;
		FillRect(BRect(x, 0, x+barSpace-1, maxHeight), B_SOLID_LOW);
		x += barSpace;
	}

	if (x<=bounds.right)
		FillRect(BRect(x,0,bounds.right,maxHeight), B_SOLID_LOW);
}

void VolumeView::UpdateVars()
{	ENTRY

	size_t bytes = sizeof(float) * 2;
	bigtime_t lc;
	volumePar->GetValue(actualGain, &bytes, &lc);
	gain[0] = TransformGain(actualGain[0]);
	gain[1] = TransformGain(actualGain[1]);
	bytes = sizeof(int32);
	mutePar->GetValue(&muted, &bytes, &lc);
}

float VolumeView::TransformGain(float x)
{	ENTRY

	switch (gainResponse) {
	case BContinuousParameter::B_LINEAR:
		return x * gainFactor + gainOffset;
	case BContinuousParameter::B_POLYNOMIAL:
		return pow(x, gainFactor) + gainOffset;
	case BContinuousParameter::B_EXPONENTIAL:
		return pow(x, gainFactor) + gainOffset;
	case BContinuousParameter::B_LOGARITHMIC:
	default:
		FatalError("Unsupported gain response");
		return 0;
	}
}

